// Window to be used with this program must have less than 255 peepholes.
// The obtained filter must be smaller than 64 MBytes.
#include <ctime>
#include <cstdio>
//#define range
#include <imgall>

typedef VETOR< VET2<int> > EE;
typedef MATRIZ<DWORD> MATDW;
typedef VET2<WORD> VETW2;

//<<<<<<<<<<<<<<<<<<< rotinas de alocacao <<<<<<<<<<<<<<<<<<<<
void *aloc(int t)
{ void *p;
  if (t==0) return NULL;
  p=malloc(t);
  if (p==NULL) erro("Out of memory at aloca");
  return p;
}

void *realoc(void *a, int t)
{ void *p;
  p=realloc(a,t);
  if (p==NULL) erro("Out of memory at realoca");
  return p;
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<< Classe FILT (cut-tree) <<<<<<<<<<<<<<<<<<<<<<<<
#define FOLHA 0xffffffff
typedef struct { DWORD v; DWORD r; } NOFILT;
// e=indice do ee (plano de corte), se for no interno.
//   valor de saida, se for folha.
// r=indice da subarvore a direita.
//   a subarvore a esquerda sempre comeca na proxima posicao do vetor filt.
//   r=FOLHA indica folha.

class FILT {
 public:
  NOFILT *v;
  int t; // Total de NOFILTs alocados.
  int n; // Quantidade de NOFILTs utilizados no momento.
  NOFILT &operator () (int i) {
    #ifdef range
    if (!(0<=i && i<n)) erro("FILT invalid index");
    #endif
    return v[i];
  }
  void aloca(int tp) { t=tp; n=0; v=(NOFILT *)aloc(sizeof(NOFILT)*t); }
  FILT(void) { t=n=0; } // default constructor
  int insere(void) {
    #ifdef range
    if (t==0) erro("FILT was not allocated");
    #endif
    if (n>=t) { t=t+t/2; v=(NOFILT *)realoc(v,sizeof(NOFILT)*t); }
    n++;
    return n-1;
  }
  //void impfilt(int fator, EE &ee, char *st);
  void lefilt(int &fator, EE &ee, char *st);
};

void FILT::lefilt(int &fator, EE &ee, char *st)
{ FILE *arq; int e;

  arq=fopen(st,"rb");
  if (arq==NULL) erro("Error in opening filter");
  e=setvbuf(arq,NULL,_IOFBF,300000);
  if (e!=0) erro("Out of memory at setvbuf");

  fread(&fator,1,sizeof(fator),arq);
  fread(&e,1,sizeof(e),arq); ee.resize(e);
  fread(ee.vet, ee.n(), sizeof(ee.vet[0]), arq);
  fread(&e, 1, sizeof(e), arq); aloca(e);
  fread(v, e, sizeof(v[0]), arq);
  n=e;
  fclose(arq);
}

//<<<<<<<<<<< Variaveis globais <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
IMGBIN ent,saibin;
MATDW sai;
EE ee;
FILT filt;
int fator;

BYTE entwee(VETW2 w, int ie)
{ return ent(w(0)+ee(ie)(0),w(1)+ee(ie)(1)); }

DWORD saiw(VETW2 w)
{ return sai(w(0),w(1)); }

//<<<<<<<<<<<< Acha/restaura <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
DWORD acha(int l, int c)
{ int i,ie; VETW2 w; BYTE b;

  w(0)=WORD(l); w(1)=WORD(c); i=0;
  while (filt(i).r!=FOLHA) {
    ie=filt(i).v;
    b=entwee(w,ie);
    if (b==0) i++; else i=filt(i).r;
  }
  return filt(i).v;
}

void restaura(void)
{ int l,c; DWORD v;
  int total,contador,umpct;

  contador=0; total=ent.nl() * ent.nc(); umpct=total/100;

  for (l=0; l<ent.nl(); l++)
    for (c=0; c<ent.nc(); c++) {

      if (contador%umpct==0) {
        fprintf(stderr,"%10.3f%% concluded...\r",100.0*contador/total);
      }
      contador++;

      v=acha(l,c);
      sai(l,c)=v;
    }
  fprintf(stderr,"%10.3f%% concluded...\r",100.0);
}

//<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
int main(int argc, char **argv)
{ int t1,t2,t3,t4,t5,i,l,c,ll,cc; int fator; DWORD d;

  if (argc!=4) {
    printf("AppZ: Zoom a binary image by applying WZ-operator constructed by BldZ\n");
    printf("APPZ To-be-zoomed.BMP Filter.CTZ Processed.BMP\n");
    erro("Error: Invalid number of arguments.");
  }
  t1=time(NULL);

  le(ent,argv[1]); ent.backg()=1;
  sai.resize(ent.nl(),ent.nc());
  filt.lefilt(fator,ee,argv[2]);
  t2=time(NULL);

  restaura();

  t3=time(NULL);
  ent.resize(0,0);
  saibin.resize(fator*sai.nl(),fator*sai.nc());
  for (l=0; l<sai.nl(); l++)
    for (c=0; c<sai.nc(); c++) {
      d=sai(l,c);
      for (ll=fator*l; ll<fator*l+fator; ll++)
        for (cc=fator*c; cc<fator*c+fator; cc++) {
          saibin(ll,cc) = d & 1;
          d=d>>1;
        }
    }

  t4=time(NULL);
  saibin.ImpBmp(argv[3]);
  t5=time(NULL);

  printf("\nZooming has finished. Zoomed image=%s\n",argv[3]);
  printf("Executed command: ");
  for (i=0; i<argc; i++) printf("%s ",argv[i]);
  printf("\n");
  printf("Total processing time: %d s\n",t4-t1);
  printf("\n\n");
}
