// Window to be used with this program must have less than 255 peepholes.
// The obtained filter must be smaller than 64 MBytes.
#include <cctype>
#include <ctime>
#include <cstring>
#include <cstdio>
//#define range
#include <imgall>

typedef VETOR< VET2<int> > EE;
typedef MATRIZ<DWORD> MATDW;
typedef VET2<WORD> VETW2;
typedef MATRIZ<double> MATDBL;
typedef VET2<int> VETI2;

//<<<<<<<<<<<<<<<<<<< rotinas de alocacao <<<<<<<<<<<<<<<<<<<<
void *aloc(int t)
{ void *p;
  if (t==0) return NULL;
  p=malloc(t);
  if (p==NULL) erro("Out of memory at aloca");
  return p;
}

void *realoc(void *a, int t)
{ void *p;
  p=realloc(a,t);
  if (p==NULL) erro("Out of memory at realoca");
  return p;
}

//<<<<<<<<<<<<<<<<<<<<<<<< classe DAT <<<<<<<<<<<<<<<<<<<<<<<<<<
class DAT { // A ultima linha de dat e' dado temporario
 public:
  DWORD *x; DWORD *y; int nl,nc,ncw;
  void aloca(int nlp, int ncp);
  void realoca(int nlp);
  DWORD &sai(int l) {
    #ifdef range
    if (!(0<=l && l<nl)) erro("DAT invalid index");
    #endif
    return y[l];
  }
  DWORD &ent(int l, int cw) {
    #ifdef range
    if (!(0<=l && l<nl)) erro("DAT iinvalid index");
    if (!(0<=cw && cw<ncw)) erro("DAT invalid index");
    #endif
    return x[ncw*l+cw];
  }
  BYTE operator () (int l, int c) {
    #ifdef range
    if (!(0<=l && l<nl)) erro("DAT invalid index");
    if (!(0<=c && c<nc)) erro("DAT invalid index");
    #endif
    return (x[ncw*l+c/32] >> (c%32)) & 1;
  }
  void put(int l, int c, BYTE v) {
    #ifdef range
    if (!(0<=l && l<nl)) erro("DAT invalid index");
    if (!(0<=c && c<nc)) erro("DAT invalid index");
    #endif
    if (v==0) x[ncw*l+c/32] &= ~(1<<(c%32));
    else      x[ncw*l+c/32] |=  (1<<(c%32));
  }
  void show(int maxl=0);
};

void DAT::aloca(int nlp, int ncp)
{ nl=nlp; nc=ncp; ncw=(nc+31)/32;
  x=(DWORD *)aloc(sizeof(DWORD)*nl*ncw);
  //for (int i=0; i<nl*ncw; i++) x[i]=0;
  y=(DWORD *)aloc(sizeof(DWORD)*nl);
  //for (int i=0; i<nl; i++) y[i]=0;
}

void DAT::realoca(int nlp)
{ nl=nlp;
  x=(DWORD *)realoc(x,sizeof(DWORD)*nl*ncw);
  if (x==NULL) erro("DAT::realoca");
  y=(DWORD *)realoc(y,sizeof(DWORD)*nl);
  if (y==NULL) erro("DAT::realoca");
}

void DAT::show(int maxl)
{ int l,c;

  if (maxl==0) maxl=nl;
  for (l=0; l<maxl; l++) {
    printf("%04d  Ent-hex:",l);
    for (c=ncw-1; c>=0; c--) printf("%08X ",unsigned((*this).ent(l,c)));
    printf("  Ent-bin:");
    for (c=nc-1; c>=0; c--) printf("%1d",(*this)(l,c));
    printf("  Sai-hex:");
    printf("%08X\n",unsigned((*this).sai(l)));
  }
  printf("\n");
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<< Classe FILT (cut-tree) <<<<<<<<<<<<<<<<<<<<<<<<
#define FOLHA 0xffffffff
typedef struct { DWORD v; DWORD r; } NOFILT;
// e=indice do ee (plano de corte), se for no interno.
//   valor de saida, se for folha.
// r=indice da subarvore a direita.
//   a subarvore a esquerda sempre comeca na proxima posicao do vetor filt.
//   r=FOLHA indica folha.

class FILT {
 public:
  NOFILT *v;
  int t; // Total de NOFILTs alocados.
  int n; // Quantidade de NOFILTs utilizados no momento.
  NOFILT &operator () (int i) {
    #ifdef range
    if (!(0<=i && i<n)) erro("FILT invalid index");
    #endif
    return v[i];
  }
  void aloca(int tp) { t=tp; n=0; v=(NOFILT *)aloc(sizeof(NOFILT)*t); }
  FILT(void) { t=n=0; } // default constructor
  int insere(void) {
    #ifdef range
    if (t==0) erro("FILT not allocated");
    #endif
    if (n>=t) { t=t+t/2; v=(NOFILT *)realoc(v,sizeof(NOFILT)*t); }
    n++;
    return n-1;
  }
  void impfilt(int fator, EE &ee, char *st);
  //void lefilt(int &fator, EE &ee, char *st);
};

void FILT::impfilt(int fator, EE &ee, char *st)
{ FILE *arq; int e;

  arq=fopen(st,"wb");
  if (arq==NULL) erro("Error in file opening");
  e=setvbuf(arq,NULL,_IOFBF,300000);
  if (e!=0) erro("Out of memory at setvbuf");

  fwrite(&fator,1,sizeof(fator),arq);
  e=ee.n();

  fwrite(&e,1,sizeof(e),arq);
  fwrite(ee.vet,ee.n(),sizeof(ee.vet[0]),arq);
  fwrite(&n,1,sizeof(n),arq);
  fwrite(v,n,sizeof(v[0]),arq);
  fclose(arq);
}

//<<<<<<<<<<< Variaveis globais <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
IMGBIN ent,saibin;
MATDW sai;
IMGGRY mat;
EE ee,eo;
FILT filt;
DAT dat;
int fator;


BYTE entwee(int l, int c, int ie)
{ return ent(l+ee(ie)(0),c+ee(ie)(1)); }

void showfilt(void)
{ int i;
  printf("****************** showfilt *********************\n");
  for (i=0; i<filt.n; i++) {
    printf("i=%d v=%u r=%u\n",i,unsigned(filt(i).v),unsigned(filt(i).r));
  }
  printf("\n");
}

//<<<<<<<<<<<<<<< ordenadat <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
int comparadat(int a, int b)
{ int i;

  for (i=dat.ncw-1; i>=0; i--) {
    if (dat.ent(a,i)<dat.ent(b,i)) return -1;
    if (dat.ent(a,i)>dat.ent(b,i)) return 1;
  }
  if (dat.sai(a)<dat.sai(b)) return -1;
  if (dat.sai(a)>dat.sai(b)) return 1;
  return 0;
}

void atribuidat(int a, int b)
{ int i;

  for (i=dat.ncw-1; i>=0; i--) {
    dat.ent(a,i)=dat.ent(b,i);
  }
  dat.sai(a)=dat.sai(b);
}

void trocadat(int a, int b)
{ int i; DWORD w;

  for (i=dat.ncw-1; i>=0; i--) {
    w=dat.ent(a,i); dat.ent(a,i)=dat.ent(b,i); dat.ent(b,i)=w;
  }
  w=dat.sai(a); dat.sai(a)=dat.sai(b); dat.sai(b)=w;
}

void ordenadat(int l, int r)
{ int i,j; static int x;
  i=l; j=r; x=dat.nl-1; atribuidat(x,(l+r)/2);
  do {
    while (comparadat(i,x)<0) i++;
    while (comparadat(x,j)<0) j--;
    if (i<=j) { trocadat(i,j); i++; j--; }
  } while (i<=j);
  if (l<j) ordenadat(l,j);
  if (i<r) ordenadat(i,r);
}

//<<<<<<<<<<<<<<< Escolhe???? <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
int comparaent(int a, int b)
{ int i;

  for (i=dat.ncw-1; i>=0; i--) {
    if (dat.ent(a,i)<dat.ent(b,i)) return -1;
    if (dat.ent(a,i)>dat.ent(b,i)) return 1;
  }
  return 0;
}

void EscolheBit(void)
{ int l,r,inserir,fator2;
  int i,j,soma[32];
  DWORD d;

  inserir=0;
  l=0; r=l;
  fator2=fator*fator;

  while (r<dat.nl-1) {
    r++;
    while (r<dat.nl-1 && comparaent(l,r)==0) r++;
    r--;

    // Processe l,r e ache media bit-a-bit
    for (j=0; j<fator2; j++) soma[j]=0;
    for (i=l; i<=r; i++) {
      d=dat.sai(i);
      for (j=0; j<fator2; j++) {
        soma[j]+=d & 1;
        d=d>>1;
      }
    }
    d=0;
    for (j=fator2-1; j>=0; j--) {
      d=d<<1;
      if (soma[j]>(r-l+1)/2) d=d+1;
    }
    dat.sai(l)=d;

    atribuidat(inserir,l); inserir++;
    l=r+1; r=l;
  }
  dat.realoca(inserir);
}

//<<<<<<<<<<<<<<< construcao <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
int diferenca(int l, int r, int p)
{ int i,br,pt;
  br=pt=0;
  for (i=l; i<=r; i++) {
    if (dat(i,p)==0) pt++; else br++;
  }
  return abs(pt-br);
}

int EixoMelhorDistribuido(int l, int r, int pee)
{ int pc,mindif,dif,i;

  mindif=diferenca(l,r,eo(pee)(0)); pc=pee;
  for (i=pee+1; i<eo.n(); i++) {
    dif=diferenca(l,r,eo(i)(0));
    if (dif<mindif) { mindif=dif; pc=i; }
  }
  return pc;
}

int partition(int l, int r, int ie)
//l..m-1:0; m..r:1 (m valor retornado)
{
  do {
    while (l<=r && dat(l,ie)==0) l++; // A esquerda de l so tem 0
    while (l<=r && dat(r,ie)==1) r--; // A direita de r so tem 1
    if (l<=r) { // Aqui, l nunca pode ser igual a r
      trocadat(l,r);
      l++; r--;
    }
  } while (l<=r);
  return l;
}

int SaidasIguais(int l, int r)
{ int i; BYTE b;
  b=dat.sai(l);
  for (i=l+1; i<=r; i++)
    if (dat.sai(i)!=b) return 0;
  return 1;
}

void trocaee(VETI2 &x, VETI2 &y)
{ VETI2 t;
  t=x; x=y; y=t;
}

void deslocaeo(int pee, int pc)
//pee<=pc
{ VETI2 t; int i;
  if (pee<pc) {
    t=eo(pee); eo(pee)=eo(pc);
    for (i=pc-1; i>pee; i--) eo(i+1)=eo(i);
    eo(pee+1)=t;
  }
}

void invdeslocaeo(int pee, int pc)
//pee<=pc
{ VETI2 t; int i;
  if (pee<pc) {
    t=eo(pee+1);
    for (i=pee+1; i<pc; i++) eo(i)=eo(i+1);
    eo(pc)=eo(pee); eo(pee)=t;
  }
}

void construcao(int l, int r, int pf, int pee)
// pf=indice de de elemento do filt vazio e previamente alocado.
// pee=indice do ee a partir do qual deve processar.
{ int pc,m,p;

  if (l==r || SaidasIguais(l,r)) { //Base da recurso
    filt(pf).v=dat.sai(l);
    filt(pf).r=FOLHA;
  } else {

    #ifdef range
    if (pee>=ee.n) erro("Isto nao deveria acontecer!");
    #endif

    pc=EixoMelhorDistribuido(l,r,pee); //pc=plano de corte
    filt(pf).v=eo(pc)(0);

    //trocaee(eo(pc),eo(pee));
    deslocaeo(pee,pc);
    m=partition(l,r,eo(pee)(0)); //l..m-1:0; m..r:1

    #ifdef range
    eo(pee,1)=0;
    #endif
    if (l<=m-1) construcao(l,m-1,filt.insere(),pee+1);

    #ifdef range
    eo(pee,1)=1;
    #endif
    p=filt.insere(); filt(pf).r=p;

    if (m<=r) construcao(m,r,p,pee+1);
    //trocaee(eo(pc),eo(pee));
    invdeslocaeo(pee,pc);
  }
}

//<<<<<<<<<<<<<<<<<< ordenaee <<<<<<<<<<<<<<<<<<<<<<<<<<
int comparaee(VETI2 a, VETI2 b)
{ int da,db;

  da=a(0)*a(0)+a(1)*a(1);
  db=b(0)*b(0)+b(1)*b(1);

  if (da<db) return -1;
  if (da>db) return 1;
  return 0;
}

void ordenaee(int l, int r)
{ int i,j; VETI2 x;
  i=l; j=r; x=ee((l+r)/2);
  do {
    while (comparaee(ee(i),x)<0) i++;
    while (comparaee(x,ee(j))<0) j--;
    if (i<=j) { trocaee(ee(i),ee(j)); i++; j--; }
  } while (i<=j);
  if (l<j) ordenaee(l,j);
  if (i<r) ordenaee(i,r);
}

//<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
int main(int argc, char **argv)
{ int l,c,nl,nc,i,ll,cc;
  int t1,t2,t3,t4,t5,t6,t7;
  DWORD d,e;

  if (argc!=5) {
    printf("BldZ: Builds WZ-operator by WZDT-learning to zoom binary image\n");
    printf("BldZ InSample.BMP OutSample.BMP Window.MAT Filter.CTZ\n");
    erro("Error: Invalid number of arguments.");
  }
  printf("Initializing...\n");

  t1=time(NULL);
  le(ent,argv[1]); ent.backg()=1;
  le(saibin,argv[2]); saibin.backg()=1;
  le(mat,argv[3]);

  t2=time(NULL);
  int conta=0;
  for (l=0; l<mat.nl(); l++)
    for (c=0; c<mat.nc(); c++)
      if (mat(l,c)!=0) conta++;
  ee.resize(conta); conta=0;
  for (l=0; l<mat.nl(); l++)
    for (c=0; c<mat.nc(); c++)
      if (mat(l,c)!=0) {
        ee(conta)(0)=l-mat.lc(); ee(conta)(1)=c-mat.cc(); conta++;
      }
  ordenaee(0,ee.n()-1);

  eo.resize(ee.n());
  for (l=0; l<ee.n(); l++) eo(l)(0)=l;

  fator=saibin.nc()/ent.nc();
  if (fator>5) erro("Error: zoom factor must be <= 5");
  if (fator*ent.nc()!=saibin.nc() || fator*ent.nl()!=saibin.nl())
    erro("Error: Incompatible image sizes.");
  nc=ent.nc(); nl=ent.nl();

  sai.resize(nl,nc);

  for (l=0; l<nl; l++)
    for (c=0; c<nc; c++) {
      d=0; i=0;
      for (ll=fator*l; ll<fator*l+fator; ll++)
        for (cc=fator*c; cc<fator*c+fator; cc++) {
          e=saibin(ll,cc);
          d+=e<<i;
          i++;
        }
      sai(l,c)=d;
    }
  saibin.resize(0,0);

  dat.aloca(nl*nc+1,ee.n()); // A ultima linha de dat e' dado temporario

  for (l=0; l<nl; l++)
    for (c=0; c<nc; c++) {
      dat.ent(l*nc+c,dat.ncw-1)=0;
      for (i=0; i<ee.n(); i++) dat.put(l*nc+c,i,entwee(l,c,i));
      dat.sai(l*nc+c)=sai(l,c);
    }
  ent.resize(0,0); sai.resize(0,0);

  t3=time(NULL);
  printf("Sorting samples...\n");
  ordenadat(0,dat.nl-2);

  //Antes de EscolheModa, dat tem uma linha a mais para temporario.
  t4=time(NULL);
  printf("Eliminating repeated samples and computing mode...\n");
  EscolheBit();

  t5=time(NULL);
  printf("Building decision tree...\n");
  filt.aloca(2*dat.nl);
  construcao(0,dat.nl-1,filt.insere(),0);

  t6=time(NULL);
  printf("Writing filter...\n");
  filt.impfilt(fator,ee,argv[4]);
  t7=time(NULL);

  printf("Executed command: ");
  for (i=0; i<argc; i++) printf("%s ",argv[i]);
  printf("\n");
  printf("Image size: rows=%d columns=%d pixels=%d\n",nl,nc,nl*nc);
  printf("Number of non-repeated samples=%d\n",dat.nl);
  printf("Number of nodes (internal+leaves)=%d.\n",filt.n);
  printf("Total processing time: %d s\n",t7-t1);
  printf("\n\n");
}

